#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to experiment procedures

# (c) Issabella (tutorial) / Dinasset (filter) / 2021
#
#   History:
#
#   v1.0: 2021-06-30: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import random
import time
import datetime
from gimpfu import *
import pygtk
import gtk
from sys import argv, platform
from ctypes import *

# define here some constant value, when required
FixWH=3000

############################################################################
# Service routines
#---------------------------------------------------------------------------

def load_library (library_name):
    if platform == "linux" or platform == "linux2":
        library_name = library_name + '.so.0'
    elif platform == "win32":
        from ctypes.util import find_library
        library_name = find_library (library_name + "-0")
    else:
        raise BaseException ("TODO")
    return CDLL (library_name)
#import sys
#print (sys.version)
gimp = load_library ('libgimp-2.0')
#print (gimp)
#gimp = load_library ('libgimp-3.0')

def gegl_graph(image, drawable, gegl_graph_string):
    class GeglBuffer(Structure):
        pass
    drawable_id = drawable.ID
    
    gegl = load_library ('libgegl-0.4')
    sucess = gegl.gegl_init (None, None)
    gimp.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
    gimp.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

    x = c_int (); y = c_int (); w = c_int (); h = c_int ()
    non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
    args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
    
    if non_empty:
        source = gimp.gimp_drawable_get_buffer (drawable_id)
        target = gimp.gimp_drawable_get_shadow_buffer (drawable_id)
        sucess = gegl.gegl_render_op (source, target, "gegl:gegl", *args)
        gegl.gegl_buffer_flush (target)
        gimp.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
        gimp.gimp_drawable_update (drawable_id, x, y, w, h)
        gimp.gimp_displays_flush ()
        
# here the single GEGL functions

def gegl_c2g(image, drawable, uradius, usamples, uiter):
    radius=int(uradius)      #dft 300   #ui_range (2, 1000)
    samples=int(usamples)    #dft 4     #ui_range (3, 17)
    iterations=int(uiter)    #dft 10    #ui_range (1, 30)
    enhance_shadows=False
    mystring="gegl:c2g radius=%d samples=%d iterations=%d enhance-shadows=%d"
    gegl_graph_string=mystring % (radius, samples, iterations, enhance_shadows)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_edge(image, drawable, ualgorithm): 
    algorithm="sobel" # 0=sobel,1=prewitt,2=gradient,3=roberts,4=differential,5=laplace
    amount=2.0        # value_range  (1.00 to 10.0)
    border=1          # 0=none,1=clamp,2=loop,3=black,4=white
    mystring="gegl:edge algorithm=%s amount=%f border=%d"
    gegl_graph_string=mystring % (algorithm, amount, border)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_median_blur (image, drawable, uradius):
    radius=float(uradius)
    percentile=50.0
    alpha_percentile=50.0
    gegl_graph_string="gegl:median-blur radius=%f percentile=%f alpha-percentile=%f" % (radius, percentile, alpha_percentile)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_gaussian_blur (image, drawable, ux, uy):    # gaussian blur  
    std_dev_x=float(ux)    #1.50     #user_range   (0.24, 100.0)
    std_dev_y=float(uy)    #1.50     #user_range   (0.24, 100.0)
    mystring="gegl:gaussian-blur std-dev-x=%f std-dev-y=%f"
    gegl_graph_string=mystring % (std_dev_x, std_dev_y)
    gegl_graph(image, drawable,gegl_graph_string)

    
############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    pdb.gimp_message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    pdb.gimp_message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    
    #pdb.gimp_selection_none (inImage)
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    optiWidthHeight = FixWH       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)
    pdb.gimp_selection_none (newImage)
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    # adjust mean values
    adjustmeanv (newImage, newBG)
    # remove low values, remove high values  
    pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                   1.00, 0.10, 0.90, False)
    newBG.opacity = 100.0
    
    return(newImage)

#---------------------------------------------------------------------------
# adjust values
#---------------------------------------------------------------------------
def adjustmeanv(inImage, inLayer):
    # new version modified to take care of the existing mean in determining the theretical mean 
    # check mean value of the layer
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    # consider 3 classes
    if meanl < 85: thmean = 100
    elif meanl > 170: thmean = 155
    else: thmean = 127.5    # regular
    minmean = thmean-16
    maxmean = thmean+16
    
    modified=False
    if meanl < minmean:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl)+" ratio = "+str(ratio))
        while meanl < minmean:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
        modified=True
        #info_msg("increased mean value = "+str(meanl))
    if meanl > maxmean:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl))
        while meanl > maxmean:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
        mofified=True
        #info_msg("reduced mean value = "+str(meanl))
    
    return(inLayer)
     
#
# ================ *
#   M  A  I  N     *
# ---------------- *
   
def JvidEffectF2 (inImage, inLayer, widthheight, drawlook, addPlasma,
                 doFlatten):
    
    # house-keeping
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # display messages using error console
    
    newImage = createOutimage (inImage, inLayer)
    Original = newImage.layers[1]
    Original.name = "Original"
    Preprocessed = newImage.layers[0]
    Preprocessed.name = "Preprocessed"
    width = Preprocessed.width
    height = Preprocessed.height
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # prepare image for detecting edges (apply median blur)
    MedianBlur = Preprocessed.copy()
    newImage.add_layer (MedianBlur, 0)
    MedianBlur.name = "Median Blurred 2.5"
    RADIUS=2.5
    gegl_median_blur (newImage, MedianBlur, RADIUS)
     
    # detect edges (apply edge (sobel)
    Sobel = MedianBlur.copy()
    newImage.add_layer (Sobel, 0)
    Sobel.name = "Edge Sobel 2"
    ALGORITHM="sobel"
    gegl_edge (newImage, Sobel, ALGORITHM)
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # invert
    SobelI = Sobel.copy()
    newImage.add_layer (SobelI, 0)
    SobelI.name = "Edge Inverted"
    pdb.gimp_drawable_invert (SobelI, False)
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # apply threshold to values
    Threshold = SobelI.copy()
    newImage.add_layer (Threshold, 0)
    Threshold.name = "Edge Threshold"
    LOW=239; HIGH=255
    pdb.gimp_threshold (Threshold, LOW, HIGH)
    pdb.gimp_item_set_visible (SobelI, False)
    pdb.gimp_item_set_visible (Sobel, False)
    Threshold.mode = LAYER_MODE_SOFTLIGHT
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # prepare first base  
    Base1 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base1, 0)
    Base1.name = "Base for Hard Paint (First Intermediate Outcome)"
    Base1.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)
    #return

    # Hard Painting  fx_hard_painting 1,2.5,2.5,1000,0,...
    # ...1,0,2,6,5,20,0,0.62,14,0,1,0.5,0.78,1.92,0,0,0,1,1,1,0.5,0.8,1.28,0,50,50
    pdb.plug_in_gmic_qt(newImage, Base1, 1,2 ,
    #[G'MIC] Hard Painting: fx_hard_painting 1,2.5,2.5,1000,0,...
        "-v - -fx_hard_painting 1,2.5,2.5,1000,0,"+
        "1,0,2,6,5,20,0,0.62,14,0,1,0.5,0.78,1.92,0,0,0,1,1,1,0.5,0.8,1.28,0,50,50 ")
    Paint = newImage.layers[0]
    Paint.name = "Hard Paint 1,2.5,2.5,1000,0"
    Paint.mode = LAYER_MODE_PIN_LIGHT
    
    #pdb.gimp_display_new (newImage)
    #return
     
    # prepare second base  
    Base2 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base2, 0)
    Base2.name = "Base for Halftone (Second Intermediate Outcome)"
    Base2.mode = LAYER_MODE_NORMAL
    if addPlasma==True:
        Plasma = Base2.copy()
        newImage.add_layer (Plasma, 0)
        pdb.plug_in_plasma (newImage, Plasma, 2, 2.0)
        pdb.gimp_drawable_curves_spline (Plasma, HISTOGRAM_VALUE,
                                         6, (0.0,0.5,0.5,0.5,1.0,1.0))
        Plasma.mode = LAYER_MODE_DARKEN_ONLY
        Base2 = pdb.gimp_image_merge_down (newImage, Plasma,
                                           CLIP_TO_BOTTOM_LAYER)
    
    #pdb.gimp_display_new (newImage)
    #return

    # Half Tone
    pdb.plug_in_gmic_qt(newImage, Base2, 1,2 ,
    #[G'MIC] HalfTone: fx_halftone 0,0,0,0,5,8,8,5,0.1,0,50,50
        "-v - -fx_halftone 0,0,0,0,5,8,8,5,0.1,0,50,50 ")
    HalfTone = newImage.layers[0]
    HalfTone.name = "HalfTone: fx_halftone 0,0,0,0,5,8,8,5,0.1,0,50,50"
    HalfTone.mode = LAYER_MODE_GRAIN_MERGE
    HalfTone.opacity = 100.0
    
    #pdb.gimp_display_new (newImage)
    #return

    Preoutcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Preoutcome, 0)
    Preoutcome.name = "Preliminary Outcome (Colours)"
    Preoutcome.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # Colors to Gray
    BWoutcome = Preoutcome.copy()
    newImage.add_layer (BWoutcome, 0)
    BWoutcome.name = "Preliminary Outcome (Black&White)"
    BWoutcome.mode = LAYER_MODE_NORMAL
    RADIUS=300; SAMPLES=4; ITER=10
    gegl_c2g (newImage, BWoutcome, RADIUS, SAMPLES, ITER)
    
    #pdb.gimp_display_new (newImage)
    #return

     
    # final step samj_Impressions 10,5,5,1,1,0,2,0,1,5,0,5,100,2,4,1,250,1.2
    if drawlook==0: #colours
        pdb.gimp_item_set_visible (BWoutcome, False)
        option = "(in colours)"
    elif drawlook==1: #B&W
        pdb.gimp_item_set_visible (Preoutcome, False)
        option = "(in B&W)"
    else: #mix
        BWoutcome.opacity = 50.0
        option = "(in mix Colours-B&W)"
    
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        outcome.name = "outcome"
        pdb.gimp_image_remove_layer (newImage, SobelI)
        pdb.gimp_image_remove_layer (newImage, Sobel)
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    outcome.name = st+" "+oname+" with JVID effect F (Halftone Graphic Art) v.2 "+option
    
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="issabella/dinasset"
menu="<Image>/Jvid/"
copyrightYear="2021"
desc="JVID Effect F2 (Halftone Graphic Art) v.2 "

register(
    "JvidEffectF2",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+" ...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+"\n (0=same as input)",
                                    3000, (0, 6000, 100)),
        (PF_OPTION, "drawlook", "Draw look", 0, ("colour","B&W","Mix")),
        (PF_TOGGLE, "plasma", "Add plasma before Half-Tone?", False),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    JvidEffectF2,
    menu=menu
)

main()
